use crate::todo::{ChangeType, TaskColumn};
use rusqlite::{params, Connection, Error, Result};
use std::path::Path;

pub(crate) fn startup(path: &Path) -> Result<(Connection)> {
    let conn = Connection::open(path)?;
    
    conn.execute(
        "CREATE TABLE IF NOT EXISTS tasks (
                id INTEGER PRIMARY KEY,
                checked INTEGER NOT NULL,
                value TEXT NOT NULL
        )",
        ()
    )?;
    
    Ok(conn)
}

pub(crate) fn get_tasks(conn: &Connection) -> std::result::Result<Vec<TaskColumn>, Error> {
    let mut stmt = conn.prepare("SELECT id, checked, value FROM tasks")?;
    let tasks_iter = stmt.query_map([], |row| {
        Ok(TaskColumn {
            id: row.get(0)?,
            checked: row.get(1)?,
            value: row.get(2)?,
        })
    })?;
    
    let mut tasks = Vec::new();
    for task in tasks_iter {
        tasks.push(task?);
    }

    Ok(tasks)
}

pub(crate) fn insert_task(conn: &Connection, task: TaskColumn) -> Result<i64> {
    conn.execute(
        "INSERT INTO tasks (checked, value) VALUES (?1, ?2)",
        (&task.checked, &task.value),
    )?;
    
    Ok(conn.last_insert_rowid())
}

pub(crate) fn update_task(conn: &Connection, task: TaskColumn, change: ChangeType) -> Result<()> {
    let (query, params) = match change {
        ChangeType::Checked => (
            "UPDATE tasks SET checked = ?1 WHERE id = ?2",
            params![&task.checked, &task.id]
        ),
        ChangeType::Value => (
            "UPDATE tasks SET value = ?1 WHERE id = ?2",
            params![&task.value, &task.id]
        ),
    };

    conn.execute(query, params)?;
    Ok(())
}

pub(crate) fn update_unselect_tasks(conn: &Connection, checked: bool) -> Result<()> {
    conn.execute(
        "UPDATE tasks SET checked = ?1",
        (&checked,)
    )?;
    
    Ok(())
}

pub(crate) fn delete_tasks(conn: &Connection, id: Option<usize>) -> Result<()> {
    let (query, params) = match id { 
        Some(t) => (
            "DELETE FROM tasks WHERE id = ?1",
            params![t.clone()]
        ),
        None => (
            "DELETE FROM tasks",
            params![]
        ),
    };

    conn.execute(query, params)?;
    Ok(())
}